/**
 * This file is part of Haketilo.
 *
 * Function: Miscellaneous operations refactored to a separate file.
 *
 * Copyright (C) 2021 Wojtek Kosior
 * Copyright (C) 2021 jahoti
 * Redistribution terms are gathered in the `copyright' file.
 */

/*
 * IMPORTS_START
 * IMPORT browser
 * IMPORT TYPE_NAME
 * IMPORT TYPE_PREFIX
 * IMPORTS_END
 */

/* Generate a random base64-encoded 128-bit sequence */
function gen_nonce()
{
    let randomData = new Uint8Array(16);
    crypto.getRandomValues(randomData);
    return btoa(String.fromCharCode.apply(null, randomData));
}

/*
 * generating unique, per-site value that can be computed synchronously
 * and is impossible to guess for a malicious website
 */

/* Uint8toHex is a separate function not exported as (a) it's useful and (b) it will be used in crypto.subtle-based digests */
function Uint8toHex(data)
{
    let returnValue = '';
    for (let byte of data)
	returnValue += ('00' + byte.toString(16)).slice(-2);
    return returnValue;
}

function gen_nonce(length=16)
{
    let randomData = new Uint8Array(length);
    crypto.getRandomValues(randomData);
    return Uint8toHex(randomData);
}

/* CSP rule that blocks scripts according to policy's needs. */
function make_csp_rule(policy)
{
    let rule = "prefetch-src 'none'; script-src-attr 'none';";
    const script_src = policy.has_payload ?
	  `'nonce-${policy.nonce}'` : "'none'";
    rule += ` script-src ${script_src}; script-src-elem ${script_src};`;
    return rule;
}

/* Check if some HTTP header might define CSP rules. */
const csp_header_regex =
      /^\s*(content-security-policy|x-webkit-csp|x-content-security-policy)/i;

/*
 * Print item together with type, e.g.
 * nice_name("s", "hello") → "hello (script)"
 */
function nice_name(prefix, name)
{
    return `${name} (${TYPE_NAME[prefix]})`;
}

/* Open settings tab with given item's editing already on. */
function open_in_settings(prefix, name)
{
    name = encodeURIComponent(name);
    const url = browser.runtime.getURL("html/options.html#" + prefix + name);
    window.open(url, "_blank");
}

/*
 * Check if url corresponds to a browser's special page (or a directory index in
 * case of `file://' protocol).
 */
const privileged_reg =
      /^(chrome(-extension)?|moz-extension):\/\/|^about:|^file:\/\/.*\/$/;
const is_privileged_url = url => privileged_reg.test(url);

/* Parse a CSP header */
function parse_csp(csp) {
    let directive, directive_array;
    let directives = {};
    for (directive of csp.split(';')) {
	directive = directive.trim();
	if (directive === '')
	    continue;

	directive_array = directive.split(/\s+/);
	directive = directive_array.shift();
	/* The "true" case should never occur; nevertheless... */
	directives[directive] = directive in directives ?
	    directives[directive].concat(directive_array) :
	    directive_array;
    }
    return directives;
}

/* Regexes and objects to use as/in schemas for parse_json_with_schema(). */
const nonempty_string_matcher = /.+/;

const matchers = {
    sha256: /^[0-9a-f]{64}$/,
    nonempty_string: nonempty_string_matcher,
    component: [
	new RegExp(`^[${TYPE_PREFIX.SCRIPT}${TYPE_PREFIX.BAG}]$`),
	nonempty_string_matcher
    ]
};

/*
 * EXPORTS_START
 * EXPORT gen_nonce
 * EXPORT make_csp_rule
 * EXPORT csp_header_regex
 * EXPORT nice_name
 * EXPORT open_in_settings
 * EXPORT is_privileged_url
 * EXPORT matchers
 * EXPORTS_END
 */
