/**
 * This file is part of Haketilo.
 *
 * Functions: Operations related to "signing" of data.
 *
 * Copyright (C) 2021 Wojtek Kosior
 * Redistribution terms are gathered in the `copyright' file.
 */

/*
 * IMPORTS_START
 * IMPORT sha256
 * IMPORT browser
 * IMPORT is_mozilla
 * IMPORTS_END
 */

/*
 * In order to make certain data synchronously accessible in certain contexts,
 * Haketilo smuggles it in string form in places like cookies, URLs and headers.
 * When using the smuggled data, we first need to make sure it isn't spoofed.
 * For that, we use this pseudo-signing mechanism.
 *
 * Despite what name suggests, no assymetric cryptography is involved, as it
 * would bring no additional benefits and would incur bigger performance
 * overhead. Instead, we hash the string data together with some secret value
 * that is supposed to be known only by this browser instance. Resulting hash
 * sum plays the role of the signature. In the hash we also include current
 * time. This way, even if signed data leaks (which shouldn't happen in the
 * first place), an attacker won't be able to re-use it indefinitely.
 *
 * The secret shared between execution contexts has to be available
 * synchronously. Under Mozilla, this is the extension's per-session id. Under
 * Chromium, this is a dummy web-accessible-resource name that resides in the
 * manifest and is supposed to be constructed by each user using a unique value
 * (this is done automatically by `build.sh').
 */

function get_secret()
{
    if (is_mozilla)
	return browser.runtime.getURL("dummy");

    return chrome.runtime.getManifest().web_accessible_resources
	.map(r => /^chromium-key-dummy-file-(.*)/.exec(r)).filter(r => r)[0][1];
}

function extract_signed(signature, signed_data)
{
    const match = /^([1-9][0-9]{12}|0)_(.*)$/.exec(signed_data);
    if (!match)
	return {fail: "bad format"};

    const result = {time: parseInt(match[1]), data: match[2]};
    if (sign_data(result.data, result.time)[0] !== signature)
	result.fail = "bad signature";

    return result;
}

/*
 * Sign a given string for a given time. Time should be either 0 or in the range
 * 10^12 <= time < 10^13.
 */
function sign_data(data, time) {
    return [sha256(get_secret() + time + data), `${time}_${data}`];
}

/*
 * EXPORTS_START
 * EXPORT extract_signed
 * EXPORT sign_data
 * EXPORTS_END
 */
